from lazyops import lazy_init
lazy_init('pydantic')
lazy_init('passlib[argon2]', 'passlib')

import abc
import sys
import threading
import time
import base64
import asyncio
import fileio


from functools import cached_property
from dataclasses import dataclass
from pydantic.typing import NoneType
from pydantic import create_model, ValidationError, validator
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar, Union
from passlib.context import CryptContext
from uuid import uuid4
from enum import Enum

from lazyops.lazyclasses import lazyclass
from lazyops.lazyrpc import BaseModel, Field, sjson_dumps, sjson_loads, jsonable_encoder
from lazyops import LazyDate, fio, get_logger, timer, tstamp, LazyEnv, LazyObject
from lazyops import timed_cache

_lazydb_default_cache_path = fio.userdir('.lazydb', mkdirs=True)
baselogger = get_logger('LazyDB')
pkler = fileio.src._pickler



from .backends import (
    LazyDBCacheBase,
    PklDBCache,
)

from .core import (
    LazyDBModel, 
    LazyDBConfig,
    LazyDBBase,
    LazyDB
)

pkldb_backend = PklDBCache
lazydb_model = LazyDBModel
lazydb_config = LazyDBConfig
lazydb = LazyDB

__all__ = [
    'LazyDBCacheBase',
    'PklDBCache',
    'LazyDBModel',
    'LazyDBConfig',
    'LazyDBBase',
    'LazyDB',
    'pkldb_backend',
    'lazydb_model',
    'lazydb_config',
    'lazydb'
]




from ._base import *

class Params(fastapi.params.Body):
    def __init__(
        self,
        default: Any,
        *,
        media_type: str = 'application/json',
        title: str = None,
        description: str = None,
        gt: float = None,
        ge: float = None,
        lt: float = None,
        le: float = None,
        min_length: int = None,
        max_length: int = None,
        regex: str = None,
        **extra: Any,
    ):
        super().__init__(
            default,
            embed=False,
            media_type=media_type,
            alias='params',
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            **extra,
        )


components = {}


def component_name(name: str, module: str = None):
    """OpenAPI components must be unique by name"""
    def decorator(obj):
        obj.__name__ = name
        obj.__qualname__ = name
        if module is not None: obj.__module__ = module  # see: pydantic.schema.get_long_model_name
        key = (obj.__name__, obj.__module__)
        if key in components:
            if components[key].schema() != obj.schema():
                raise RuntimeError(f"Different models with the same name detected: {obj!r} != {components[key]}")
            return components[key]
        components[key] = obj
        return obj
    return decorator


def is_scope_child(owner: type, child: type):
    return (
        (
            owner.__dict__.get(child.__name__) is child or
            owner.__dict__.get(child.__name__) is Optional[child]
        ) and
        child.__qualname__ == owner.__qualname__ + '.' + child.__name__ and
        child.__module__ == owner.__module__
    )


def rename_if_scope_child_component(owner: type, child, postfix: str):
    if is_scope_child(owner, child): child = component_name(f'{owner.__name__}.{postfix}', owner.__module__)(child)
    return child



import requests
import aiohttp
from lazyops import async_cache, timed_cache
from lazyops.lazyio import LazyJson
from ._base import lazyclass, dataclass, List, Union, Any
from .tfserving_pb2 import TFSModelConfig, TFSConfig



@lazyclass
@dataclass
class TFSRequest:
    data: Any

    def to_data(self):
        if isinstance(self.data, str):
            return {'inputs': [self.data]}
        if isinstance(self.data, list):
            return {'inputs': self.data}
        if isinstance(self.data, dict):
            if self.data.get('inputs'):
                return self.data
            if self.data.get('text'):
                if isinstance(self.data['text'], str):
                    return {'inputs': [self.data['text']]}
                return {'inputs': self.data['text']}
        return {'inputs': self.data}


@lazyclass
@dataclass
class TFSModelEndpoint:
    url: str
    config: TFSModelConfig
    ver: str = 'v1'
    sess: requests.Session = requests.Session()
    aiosess: aiohttp.ClientSession = None

    @property
    def default_endpoint(self):
        return f'{self.url}/{self.ver}/models/{self.config.name}'
    
    @property
    def default_url(self):
        if self.config.default_label:
            return self.default_endpoint + f'/labels/{self.config.default_label}'
        if self.config.default_version:
            return self.default_endpoint + f'/versions/{self.config.default_version}'
        return self.default_endpoint
    
    @property
    def default_predict(self):
        return self.default_url + ':predict'
    
    def get_label(self, label: str = 'latest'):
        return self.default_endpoint + '/labels/' + label
    
    def get_version(self, ver: Union[str, int]):
        return self.default_endpoint + '/versions/' + str(ver)
    
    def get_endpoint(self, label: str = None, ver: Union[str, int] = None):
        if label:
            return self.get_label(label)
        if ver:
            return self.get_version(ver)
        return self.default_url
    
    def get_predict_endpoint(self, label: str = None, ver: Union[str, int] = None):
        if label:
            return self.get_label(label) + ':predict'
        if ver:
            return self.get_version(ver) + ':predict'
        return self.default_predict
        
    @timed_cache(seconds=600)
    def get_metadata(self, label: str = None, ver: Union[str, int] = None):
        ep = self.get_endpoint(label=label, ver=ver)
        return self.sess.get(ep + '/metadata').json()
    
    @async_cache
    async def get_aio_metadata(self, label: str = None, ver: Union[str, int] = None):
        ep = self.get_endpoint(label=label, ver=ver)
        resp = await self.aiosess.get(ep + '/metadata')
        return resp.json()

    @async_cache
    async def aio_predict(self, data, label: str = None, ver: Union[str, int] = None, **kwargs):
        epoint = self.get_predict_endpoint(label=label, ver=ver)
        item = TFSRequest(data=data)
        res = await self.aiosess.post(epoint, json=item.to_data())
        return res.json(loads=LazyJson.loads)
    
    @timed_cache(seconds=60)
    def predict(self, data, label: str = None, ver: Union[str, int] = None, **kwargs):
        epoint = self.get_predict_endpoint(label=label, ver=ver)
        item = TFSRequest(data=data)
        res = self.sess.post(epoint, json=item.to_data())
        return res.json(loads=LazyJson.loads)
    
    @property
    def is_alive(self):
        return bool(self.get_metadata().get('model_spec'))
    

class TFServeModel:
    def __init__(self, url: str, configs: List[TFSModelConfig], ver: str = 'v1', **kwargs):
        self.url = url
        self.configs = configs
        self.ver = ver
        self.sess = requests.Session()
        self.aiosess = aiohttp.ClientSession()
        self.endpoints = {config.name: TFSModelEndpoint(url, config=config, ver=ver, sess=self.sess, aiosess=self.aiosess) for config in configs}
        self.default_model = kwargs.get('default_model') or configs[0].name
        self.available_models = [config.name for config in configs]

    def predict(self, data, model: str = None, label: str = None, ver: Union[str, int] = None, **kwargs):
        if model: assert model in self.available_models
        model = model or self.default_model
        return self.endpoints[model].predict(data, label=label, ver=ver, **kwargs)

    async def aio_predict(self, data, model: str = None, label: str = None, ver: Union[str, int] = None, **kwargs):
        if model: assert model in self.available_models
        model = model or self.default_model
        return await self.endpoints[model].aio_predict(data, label=label, ver=ver, **kwargs)
    
    @property
    def api_status(self):
        return {model: self.endpoints[model].is_alive for model in self.available_models}
    
    @timed_cache(seconds=600)
    def get_api_metadata(self):
        return {model: self.endpoints[model].get_metadata() for model in self.available_models}

    @async_cache
    async def get_aio_api_metadata(self):
        return {model: await self.endpoints[model].get_aio_metadata() for model in self.available_models}
    

    @classmethod
    def from_config_file(cls, url, filepath: str, ver: str = 'v1', **kwargs):
        configs = TFSConfig.from_config_file(filepath, as_obj=True)
        return TFServeModel(url=url, configs=configs, ver=ver, **kwargs)
    



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import subprocess
import importlib
import pkg_resources
import threading
from subprocess import check_output
from dataclasses import dataclass
from typing import Optional
from fileio import File, PathIO, PathIOLike
from fileio.src import get_pathlike, autojson, read_json, read_jsonlines


from lazyops.envs import logger
from lazyops.envs import LazyEnv



def lazy_import(name):
    if name in sys.modules:
        return sys.modules[name]
    sys.modules[name] = importlib.import_module(name)
    return sys.modules[name]

def exec_command(cmd):
    out = check_output(cmd, shell=True)
    if isinstance(out, bytes): out = out.decode('utf8')
    return out.strip()

def run_cmd(cmd):
    return exec_command(cmd)

def lazy_check(req):
    try:
        _ = pkg_resources.get_distribution(req)
        return True
    except pkg_resources.DistributionNotFound:
        return False


def lazy_install(req, force=False, latest=False, verbose=False):
    req_base = req.split('=')[0].replace('>','').replace('<','').strip()
    if lazy_check(req_base) and not force:
        return
    python = sys.executable
    pip_exec = [python, '-m', 'pip', 'install']
    if '=' not in req or latest:
        pip_exec.append('--upgrade')
    pip_exec.append(req)
    subprocess.check_call(pip_exec, stdout=subprocess.DEVNULL)
    if verbose:
        logger.info(f'{req} installed successfully.')

def lazy_init(name, lib_name=None, force=False, latest=False, verbose=False):
    lazy_install(name, force=force, latest=latest, verbose=verbose)
    return lazy_import(lib_name or name)


def clone_repo(repo, path=None, absl=False, add_to_syspath=True):
    path = path or ('/content' if LazyEnv.is_colab else File.curdir())
    if isinstance(repo, str): repo = repo.split(',')
    assert isinstance(repo, list), f'Repo must be a list or string: {type(repo)}'
    logger.info(f'Cloning Repo(s): {repo} into {path}')
    for r in repo:
        if 'github.com' not in r:
            r = f'https://github.com/{r}'
        clonepath = File.join(path, File.base(r)) if not absl else path
        try:
            run_cmd(f'git clone {r} {clonepath}')
        except Exception as e:
            logger.error(f'Error Cloning {r}: {str(e)}')
        if add_to_syspath:
            sys.path.append(clonepath)


@dataclass
class LazySubmodule:
    name: str # 'models'
    repo: Optional[str] = None
    module: Optional[str] = None # 'libname'
    module_path: Optional[str] = None # 'libname.nested' = 'libname.nested.models'
    init: Optional[bool] = False

    @property
    def main_module(self):
        return importlib.util.find_spec(self.name)
    
    @property
    def main_module_path(self):
        if not self.main_module:
            return None
        return self.main_module.submodule_search_locations[0]

    @property
    def submodule_name(self):
        if not self.module:
            return self.name
        if not self.module_path:
            return self.module + '.' + self.name
        if self.name in self.module_path:
            return self.module_path
        return self.module_path + '.' + self.name
    
    @property
    def submodule_namepath(self):
        return '/'.join(self.submodule_name.split('.'))

    @property
    def submodule_path(self):
        if not self.main_module_path:
            return None
        return File.join(self.main_module_path, self.submodule_namepath)
    
    @property
    def has_initialized(self):
        return bool(self.submodule_name in sys.modules)

    def lazy_import(self):
        if self.submodule_name and self.has_initialized:
            return
        if not File.exists(self.submodule_path) and self.repo and self.init:
            clone_repo(self.repo, path=self.submodule_path, abls=True)
        if File.exists(self.submodule_path):
            sys.modules[self.submodule_name] = importlib.import_module(self.submodule_name)
            logger.info(f'Initialized Submodule: {self.submodule_name}')
        else:
            logger.warn(f'Failed to Initilize Submodule: {self.submodule_name}')



class LazyImporter:
    libs = {}
    submodules = {}
    def __init__(self):
        self.lock = threading.RLock()
    
    @property
    def imports(self):
        return LazyImporter.libs
    
    def has_submodule(name, *args, **kwargs):
        bool(LazyImporter.get_submodule(name)) 

    def get_submodule(name, *args, **kwargs):
        return LazyImporter.submodules.get(name, None)

    def setup_submodule(self, name, *args, **kwargs):
        self.lock.acquire()
        with self.lock:
            if name not in LazyImporter.submodules:
                LazyImporter.submodules[name] = LazySubmodule(name=name, *args, **kwargs)
                LazyImporter.submodules[name].lazy_init()
            return LazyImporter.submodules[name]

    def setup_lib(self, name, lib_name=None, force=False, latest=False, verbose=False):
        self.lock.acquire()
        with self.lock:
            if name not in LazyImporter.libs:
                LazyImporter.libs[name] = lazy_init(name, lib_name, force, latest, verbose)
            return LazyImporter.libs[name]
    
    def has_lib(self, name, lib_name=None):
        return bool(self.get_lib(name, lib_name))

    def get_lib(self, name, lib_name=None):
        return LazyImporter.libs.get(name, LazyImporter.libs.get(lib_name, None))

    def __call__(self, name, lib_name=None, *args, **kwargs):
        if self.has_lib(name, lib_name):
            return self.get_lib(name, lib_name)
        return self.setup_lib(name, lib_name=lib_name, *args, **kwargs)

lazylibs = LazyImporter()





from ._base import *

from .filters import GDELTFilters
from .models import GDELTArticle

class GDELTFormat(Enum):
    dict = 'dict'
    obj = 'obj'
    json = 'json'
    pandas = 'pd'


class GDELTMethods(Enum):
    article = 'article'
    timeline = 'timeline'

class GDELT:
    api_url = 'https://api.gdeltproject.org/api/v2/doc/doc'
    available_modes = ["artlist", "timelinevol", "timelinevolraw", "timelinetone", "timelinelang", "timelinesourcecountry"]

    def __init__(self, result_format: GDELTFormat = GDELTFormat.obj, json_parsing_max_depth: int = 100, *args, **kwargs) -> None:
        if isinstance(result_format, str): result_format = GDELTFormat[result_format]
        self.max_depth_json_parsing = json_parsing_max_depth
        self._output_format = result_format
        self.sess = LazySession()

    def return_article_result(self, articles: Dict = None):
        if not articles or not articles.get('articles'):
            return None
        if self._output_format.value == 'dict':
            return articles['articles']
        
        if self._output_format.value == 'pd':
            return pd.DataFrame(articles["articles"])
        
        if self._output_format.value == 'json':
            return LazyJson.dumps(articles['articles'])
        
        if self._output_format.value == 'obj':
            return [GDELTArticle(**article) for article in articles['articles']]
    
    def return_timeline_search(self, results: Dict = None):
        if not results:
            return None
        
        if self._output_format.value == 'dict':
            return results
        
        if self._output_format.value == 'pd':
            formatted = pd.DataFrame(results)
            formatted["datetime"] = pd.to_datetime(formatted["datetime"])
            return formatted
        
        if self._output_format.value == 'json':
            return LazyJson.dumps(results)
        
        if self._output_format.value == 'obj':
            return [LazyObject(res) for res in results]


    def article_search(self, filters: GDELTFilters) -> Union[pd.DataFrame, Dict, str]:
        articles = self._query("artlist", filters.query_string)
        return self.return_article_result(articles)

    def timeline_search(self, mode: str, filters: GDELTFilters) -> Union[pd.DataFrame, Dict, str]:
        timeline = self._query(mode, filters.query_string)
        results = {"datetime": [entry["date"] for entry in timeline["timeline"][0]["data"]]}
        for series in timeline["timeline"]:
            results[series["series"]] = [entry["value"] for entry in series["data"]]

        if mode == "timelinevolraw": results["All Articles"] = [entry["norm"] for entry in timeline["timeline"][0]["data"]]
        return self.return_timeline_search(results)
    
    def search(self, method: GDELTMethods, filters: GDELTFilters) -> Union[pd.DataFrame, Dict, str]:
        if method.value == 'article':
            return self.article_search(filters)
        if method.value == 'timeline':
            return self.timeline_search(filters)

    async def async_search(self, method: GDELTMethods, filters: GDELTFilters) -> Union[pd.DataFrame, Dict, str]:
        if method.value == 'article':
            return await self.async_article_search(filters)
        if method.value == 'timeline':
            return await self.async_timeline_search(filters)

    async def async_article_search(self, filters: GDELTFilters) -> Union[pd.DataFrame, Dict, str]:
        articles = await self._async_query("artlist", filters.query_string)
        return self.return_article_result(articles)

    async def async_timeline_search(self, mode: str, filters: GDELTFilters) -> Union[pd.DataFrame, Dict, str]:
        timeline = await self._async_query(mode, filters.query_string)
        results = {"datetime": [entry["date"] for entry in timeline["timeline"][0]["data"]]}
        for series in timeline["timeline"]:
            results[series["series"]] = [entry["value"] for entry in series["data"]]

        if mode == "timelinevolraw": results["All Articles"] = [entry["norm"] for entry in timeline["timeline"][0]["data"]]
        return self.return_timeline_search(results)
    
    def _decode_json(cls, content, max_recursion_depth: int = 100, recursion_depth: int = 0):
        try:
            result = LazyJson.loads(content, recursive=True)
        except Exception as e:
            if recursion_depth >= max_recursion_depth:
                raise ValueError("Max Recursion depth is reached. JSON can´t be parsed!")
            idx_to_replace = int(e.pos)
            if isinstance(content, bytes): content.decode("utf-8")
            json_message = list(content)
            json_message[idx_to_replace] = ' '
            new_message = ''.join(str(m) for m in json_message)
            return GDELT._decode_json(content=new_message, max_recursion_depth=max_recursion_depth, recursion_depth=recursion_depth+1)
        return result
    
    async def _async_decode_json(cls, content, max_recursion_depth: int = 100, recursion_depth: int = 0):
        try:
            result = LazyJson.loads(content, recursive=True)
        except Exception as e:
            if recursion_depth >= max_recursion_depth:
                raise ValueError("Max Recursion depth is reached. JSON can´t be parsed!")
            idx_to_replace = int(e.pos)
            if isinstance(content, bytes): content.decode("utf-8")
            json_message = list(content)
            json_message[idx_to_replace] = ' '
            new_message = ''.join(str(m) for m in json_message)
            return await GDELT._async_decode_json(content=new_message, max_recursion_depth=max_recursion_depth, recursion_depth=recursion_depth+1)
        return result

    def _query(self, mode: str, query_string: str) -> Dict:
        if mode not in GDELT.available_modes:
            raise ValueError(f"Mode {mode} not in supported API modes")
        resp = self.sess.fetch(url=GDELT.api_url, decode_json=False, method='GET', params={'query': query_string, 'mode': mode, 'format': 'json'})
        if resp.status_code not in [200, 202]:
            raise ValueError("The gdelt api returned a non-successful status code. This is the response message: {}".format(resp.text))
        return self._decode_json(resp.content, max_recursion_depth=self.max_depth_json_parsing)
    
    async def _async_query(self, mode: str, query_string: str) -> Dict:
        if mode not in GDELT.available_modes:
            raise ValueError(f"Mode {mode} not in supported API modes")
        resp = await self.sess.async_fetch(url=GDELT.api_url, decode_json=False, method='GET', params={'query': query_string, 'mode': mode, 'format': 'json'})
        if resp.status_code not in [200, 202]:
            raise ValueError("The gdelt api returned a non-successful status code. This is the response message: {}".format(resp.text))
        return await self._async_decode_json(resp.content, max_recursion_depth=self.max_depth_json_parsing)




import os
import asyncio
from types import coroutine
from fileio import PathIO
from io import (
    FileIO,
    TextIOBase,
    BufferedReader,
    BufferedWriter,
    BufferedRandom,
)
from functools import partial, singledispatch, wraps
from typing import TypeVar, Union
from aiofiles.threadpool.binary import AsyncBufferedIOBase, AsyncBufferedReader, AsyncFileIO
from aiofiles.threadpool.text import AsyncTextIOWrapper
from aiofiles.base import AiofilesContextManager
from tensorflow.python.platform.gfile import GFile
from tensorflow.python.lib.io import file_io as tfio
from tensorflow.python.lib.io import _pywrap_file_io
#io import stat, rename, create_dir, remove

def async_open(
    file,
    mode="r",
    buffering=-1,
    encoding=None,
    errors=None,
    newline=None,
    closefd=True,
    opener=None,
    *,
    loop=None,
    executor=None
):
    return AiofilesContextManager(
        _open(
            file,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            closefd=closefd,
            opener=opener,
            loop=loop,
            executor=executor,
        )
    )


@coroutine
def _open(
    file,
    mode="r",
    buffering=-1,
    encoding=None,
    errors=None,
    newline=None,
    closefd=True,
    opener=None,
    *,
    loop=None,
    executor=None
):
    """Open an asyncio file."""
    if loop is None:
        loop = asyncio.get_event_loop()
    pathio = PathIO(file)
    cb = partial(
        pathio.open,
        #file,
        mode=mode,
        #buffering=buffering,
        encoding=encoding,
        #errors=errors,
        #newline=newline,
        #closefd=closefd,
        #opener=opener,
    )
    f = yield from loop.run_in_executor(executor, cb)

    return wrap(f, loop=loop, executor=executor)


@singledispatch
def wrap(file, *, loop=None, executor=None):
    raise TypeError("Unsupported io type: {}.".format(file))


@wrap.register(TextIOBase)
def _(file, *, loop=None, executor=None):
    return AsyncTextIOWrapper(file, loop=loop, executor=executor)


@wrap.register(BufferedWriter)
def _(file, *, loop=None, executor=None):
    return AsyncBufferedIOBase(file, loop=loop, executor=executor)


@wrap.register(BufferedReader)
@wrap.register(BufferedRandom)
def _(file, *, loop=None, executor=None):
    return AsyncBufferedReader(file, loop=loop, executor=executor)


@wrap.register(FileIO)
def _(file, *, loop=None, executor=None):
    return AsyncFileIO(file, loop, executor)


@wrap.register(GFile)
def _(file, *, loop=None, executor=None):
    return AsyncFileIO(file, loop, executor)


def pathlike_wrap(func):
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)

    return run


aio_stat = pathlike_wrap(tfio.stat)
aio_rename = pathlike_wrap(tfio.rename)
aio_remove = pathlike_wrap(tfio.delete_file)
aio_mkdir = pathlike_wrap(tfio.create_dir)
aio_rmdir = pathlike_wrap(tfio.delete_recursively)

stat_results = TypeVar("stat_results", str, Union[_pywrap_file_io.FileStatistics, os.stat_result])


from .core import GDELT
from .filters import GDELTFilters, near, repeat, multi_repeat
from .models import GDELTArticle, Article

__all__ = [
    'GDELT',
    'GDELTFilters',
    'near',
    'repeat',
    'multi_repeat',
    'GDELTArticle',
    'Article'
]


import os
import sys
import signal
import threading
import logging
from dataclasses import dataclass
from functools import partialmethod
from typing import Optional, Any, Dict, List
from .mp_utils import _CPU_CORES, _MAX_THREADS, _MAX_PROCS
try:
    from google.colab import auth
    _colab = True
except ImportError:
    _colab = False

_notebook = sys.argv[-1].endswith('json')

_logging_levels = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warn': logging.WARN,
    'warning': logging.WARNING,
    'error': logging.ERROR,
}
# Ignore Some Loggers
logging.getLogger('aiocache').setLevel(logging.ERROR)


class ThreadSafeHandler:
    def __init__(
        self, 
        lock_mode: str = 'lock',
        handler: Optional[Any] = None):
        self.lock_mode = lock_mode
        self.handler = handler
        self.lock = threading.RLock()  if self.lock_mode == 'rlock' else threading.Lock()

    def get(self, init_func: Any = None, *args, **kwargs):
        if self.lock_mode == 'rlock': self.lock.acquire()
        with self.lock:
            if not self.handler:
                self.handler = init_func(*args, **kwargs)
            return self.handler


class LogFormatter(logging.Formatter):
    COLOR_CODES = {
        logging.CRITICAL: "\033[38;5;196m", # bright/bold magenta
        logging.ERROR:    "\033[38;5;9m", # bright/bold red
        logging.WARNING:  "\033[38;5;11m", # bright/bold yellow
        logging.INFO:     "\033[38;5;111m", # white / light gray
        logging.DEBUG:    "\033[1;30m"  # bright/bold black / dark gray
    }

    RESET_CODE = "\033[0m"
    def __init__(self, color, *args, **kwargs):
        super(LogFormatter, self).__init__(*args, **kwargs)
        self.color = color

    def format(self, record, *args, **kwargs):
        if (self.color == True and record.levelno in self.COLOR_CODES):
            record.color_on  = self.COLOR_CODES[record.levelno]
            record.color_off = self.RESET_CODE
        else:
            record.color_on  = ""
            record.color_off = ""
        return super(LogFormatter, self).format(record, *args, **kwargs)

class LazyOpsLogger:
    def __init__(self, config):
        self.config = config
        self.logger = self.setup_logging()
    
    def setup_logging(self):
        logger = logging.getLogger(self.config['name'])
        logger.setLevel(_logging_levels[self.config.get('log_level', 'info')])

        console_log_output = sys.stdout if _notebook or _colab else sys.stderr        
        console_handler = logging.StreamHandler(console_log_output)
        console_handler.setLevel(self.config["console_log_level"].upper())
        console_formatter = LogFormatter(fmt=self.config["log_line_template"], color=self.config["console_log_color"])
        console_handler.setFormatter(console_formatter)
        if self.config.get('clear_handlers', False) and logger.hasHandlers():
            logger.handlers.clear()
        logger.addHandler(console_handler)
        if self.config.get('quiet_loggers'):
            to_quiet = self.config['quiet_loggers']
            if isinstance(to_quiet, str): to_quiet = [to_quiet]
            for clr in to_quiet:
                clr_logger = logging.getLogger(clr)
                clr_logger.setLevel(logging.ERROR)

        logger.propagate = self.config.get('propagate', False)
        return logger

    def get_logger(self, module=None):
        if module:
            return self.logger.getChild(module)
        return self.logger
    
    def debug(self, msg, *args, **kwargs):
        return self.logger.debug(msg, *args, **kwargs)
    
    def info(self, msg, *args, **kwargs):
        return self.logger.info(msg, *args, **kwargs)
    
    def warn(self, msg, *args, **kwargs):
        return self.logger.warn(msg, *args, **kwargs)
    
    def error(self, msg, *args, **kwargs):
        return self.logger.error(msg, *args, **kwargs)
    
    def exception(self, msg, *args, **kwargs):
        return self.logger.exception(msg, *args, **kwargs)
    
    def __call__(self, msg, *args, **kwargs):
        return self.logger.info(msg, *args, **kwargs)


def setup_new_logger(name, log_level='info', quiet_loggers=None, clear_handlers=False, propagate=True):
    logger_config = {
        'name': name,
        'log_level': log_level,
        'console_log_output': "stdout", 
        'console_log_level': "info",
        'console_log_color': True,
        'logfile_file': None,
        'logfile_log_level': "debug",
        'logfile_log_color': False,
        'log_line_template': f"%(color_on)s[{name}] %(name)-5s %(funcName)-5s%(color_off)s: %(message)s",
        'clear_handlers': clear_handlers,
        'quiet_loggers': quiet_loggers,
        'propagate': propagate
    }
    return LazyOpsLogger(logger_config)

IsLazyAlive = True

class EnvChecker:
    is_colab = _colab
    is_notebook = _notebook
    is_jpy = _notebook
    cpu_cores = _CPU_CORES
    max_threads = _MAX_THREADS
    max_procs = _MAX_PROCS
    loggers: Optional[Dict[str, ThreadSafeHandler]] = {}
    handlers: Optional[Dict[str, Any]] = {}
    watcher_enabled: Optional[bool] = False
    threads: Optional[List[threading.Thread]] = []
    sigs: Optional[Dict[str, signal.signal]] = {}
    
    
    @classmethod
    def get_logger(cls, name = 'LazyOps', module=None, *args, **kwargs):
        if not EnvChecker.loggers.get(name):
            EnvChecker.loggers[name] = ThreadSafeHandler()
        if module:
            return EnvChecker.loggers[name].get(setup_new_logger, name=name, *args, **kwargs).get_logger(module=module)
        return EnvChecker.loggers[name].get(setup_new_logger, name=name, *args, **kwargs)
    
    @property
    def alive(self):
        return IsLazyAlive
    
    @property
    def killed(self):
        return not IsLazyAlive
    
    @property
    def is_threadsafe(self):
        return bool(threading.current_thread() is threading.main_thread())
    
    @classmethod
    def set_state(cls, state: bool):
        global IsLazyAlive
        IsLazyAlive = state
    

    set_alive = partialmethod(set_state, True)
    set_dead = partialmethod(set_state, False)

    @classmethod
    def exit_handler(cls, signum, frame):
        logger = EnvChecker.loggers['LazyWatch']
        logger.error("Received SIGINT or SIGTERM! Gracefully Exiting.")
        if EnvChecker.handlers:
            for handler, func in EnvChecker.handlers.items():
                logger.error(f'Calling Exit for {handler}')
                func()
        if EnvChecker.threads:
            for thread in EnvChecker.threads:
                thread.join()
        EnvChecker.set_dead()
        sys.exit(0)

    @classmethod
    def enable_watcher(cls):
        if EnvChecker.watcher_enabled:
            return
        EnvChecker.sigs['sigint'] = signal.signal(signal.SIGINT, EnvChecker.exit_handler)
        EnvChecker.sigs['sigterm'] = signal.signal(signal.SIGTERM, EnvChecker.exit_handler)
        EnvChecker.loggers['LazyWatch'] = EnvChecker.get_logger(name='LazyWatch')
        EnvChecker.watcher_enabled = True

    @classmethod
    def add_thread(cls, t):
        EnvChecker.threads.append(t)

    @classmethod
    def set_multiparams(cls, max_procs: int = None, max_threads: int = None):
        if max_procs is not None: EnvChecker.max_procs = max_procs
        if max_threads is not None: EnvChecker.max_threads = max_threads

    @classmethod
    def add_exit_handler(cls, name, func):
        if name not in EnvChecker.handlers:
            EnvChecker.handlers[name] = func


    @classmethod
    def getset(cls, name, val=None, default=None, set_if_none=False):
        eval = os.environ.get(name, default=default)
        if (not eval and set_if_none) or (eval and eval != val):
            os.environ[name] = str(val)
            return os.environ.get(name)
        if not val or not set_if_none:
            return eval
        return eval

    def __call__(self, name, val=None, default=None, set_if_none=None, *args, **kwargs):
        return EnvChecker.getset(name, val=None, default=None, set_if_none=None, *args, **kwargs)
    
    def __exit__(self, type, value, traceback, *args, **kwargs):
        if EnvChecker.watcher_enabled:
            if self.killed:
                sys.exit(0)
            signal.signal(signal.SIGINT, EnvChecker.sigs['sigint'])
            signal.signal(signal.SIGTERM, EnvChecker.sigs['sigterm'])
        EnvChecker.set_dead()


    def watch(self):
        self.enable_watcher()
        return self

        

LazyEnv = EnvChecker()
get_logger = LazyEnv.get_logger
logger = get_logger()
lazywatcher = LazyEnv.watch


def lazywatch(name, *args, **kwargs):
    def decorator(func):
        def wrapper(*args, **kwargs):
            logger.info(f'Adding {name} to Exit Handlers')
            LazyEnv.add_exit_handler(name, func)
            return func(*args, **kwargs)
        return wrapper
    return decorator


from ._base import *
import operator as op

logger = get_logger('LazyDB', module='models')

# Base Class for the File-backed LazyDB Backend.
class LazyDBCacheBase(abc.ABC):
    def __init__(self, save_path=None, cache_name='lazycache', *args, **kwargs):
        self._cachename = cache_name
        self.save_path = save_path or _lazydb_default_cache_path
        fio.mkdirs(self.save_path)
        self.lock = threading.RLock()
        self._args = args
        self._kwargs = kwargs

    @property
    def cache_file(self):
        return self._cachename + '.lazydb'

    @property
    def cache_filepath(self):
        return fio.join(self.save_path, self.cache_file)
    
    @property
    def exists(self):
        return fio.exists(self.cache_filepath)

    @abc.abstractmethod
    def dumps(self, data, *args, **kwargs):
        pass
    
    @abc.abstractmethod
    def loads(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def restore(self, *args, **kwargs):
        pass
    
    @abc.abstractmethod
    def save(self, db, *args, **kwargs):
        pass

@lazyclass
@dataclass
class LazyDBConfig:
    dbschema: Dict[str, Any]
    autosave: bool = True
    autouser: bool = True
    is_dev: bool = True
    savefreq: float = 15.0
    seeddata: Optional[Dict[str, Any]] = None
    userconfigs: Optional[Dict[str, Any]] = None
    hashschema: Optional[Dict[str, Any]] = None
    dbname: Optional[str] = None
    

@dataclass
class LazyDBSaveMetrics:
    created: Optional[str] = None
    num_saved: Optional[int] = 0
    num_loaded: Optional[int] = 0
    last_load: Optional[str] = None
    last_save: Optional[str] = None
    time_alive: Optional[float] = 0


class LazyAny(object):
    def __init__(self, data):
        for k,v in data.items():
            setattr(self, k, v)


class LazyHasher:
    hasher = CryptContext(schemes=["argon2"], deprecated="auto")
    
    @classmethod
    def create(cls, pass_string: str):
        return LazyHasher.hasher.hash(pass_string)

    @classmethod
    def verify(cls, pass_hash: str, pass_string: str):
        return LazyHasher.hasher.verify(pass_string, pass_hash)
    
    @classmethod
    def update(cls, old_hash: str, old_pass: str, new_pass: str, do_verify: bool = True):
        if (do_verify and LazyHasher.verify(old_hash, old_pass)) or not do_verify:
            return LazyHasher.create(new_pass)
        return None
    
    @classmethod
    def create_token(cls):
        return str(uuid4())


class LazyUserSchema(BaseModel):
    role: str = ...
    is_super: bool = False
    username: str = ...
    password: str = ...
    hash_password: str = None

    @classmethod
    def get_schema(cls, config: Dict[str, Any] = None, is_dev: bool = True, *args, **kwargs):
        current_schema = LazyUserSchema.__fields__
        schema_data = {
            field: (vals.type_, ...)
            if vals.required
            else (vals.type_, vals.default)
            for field, vals in current_schema.items()
        }
        if not is_dev: schema_data.pop('password')
        if not config:
            return schema_data
        new_schema = config if isinstance(config, dict) else config.dict(exclude_unset=True)
        for field, values in new_schema.items():
            schema_data[field] = values
        return schema_data
    
    @classmethod
    def get_hash_schema(cls):
        return {'password': 'hash_password'}

# IS = a [val] == b [val]
# NOT = a [val] != b [val]
# IN = a [val, dict.key] in b [list, dict.keys]
# HAS = a [list, dict.values] has b [val, dict.key]


class LazyLink(Enum):
    IS = 0
    NOT = 1
    IN = 2
    HAS = 3


# 

class LazyDBLink:
    def __init__(
        self,
        name: str,
        index,
        method: Optional[LazyLink] = LazyLink.HAS,
        *args, **kwargs):
        pass

    def __ne__(self, o: object) -> bool:
        pass

    def __eq__(self, o: object) -> bool:
        pass

    def __ge__(self, o):
        pass

    def __gt__(self, o):
        pass

    def __lt__(self, o):
        pass

    def __le__(self, o):
        pass




# Has all the base dependencies

from lazyops import lazy_init, get_logger

lazy_init('fastapi')
lazy_init('pydantic')
lazy_init('aiojobs')
lazy_init('starlette')

import asyncio
import inspect
import contextvars

from collections import ChainMap
from json import JSONDecodeError
from types import FunctionType, CoroutineType
from typing import List, Union, Any, Callable, Type, Optional, Dict, Sequence, Awaitable, Tuple
from contextlib import AsyncExitStack, AbstractAsyncContextManager, asynccontextmanager, contextmanager

# noinspection PyProtectedMember
from pydantic import DictError
from pydantic import StrictStr, ValidationError
from pydantic import BaseModel as PyBaseModel
from pydantic import BaseConfig
from pydantic.fields import ModelField, Field
from pydantic.main import ModelMetaclass

# noinspection PyProtectedMember
from fastapi.dependencies.models import Dependant
from fastapi.encoders import jsonable_encoder
from fastapi.params import Depends
from fastapi import FastAPI, Body, Header
from fastapi.dependencies.utils import solve_dependencies, get_dependant, get_flat_dependant, get_parameterless_sub_dependant
from fastapi.exceptions import RequestValidationError, HTTPException
from fastapi.routing import APIRoute, APIRouter, serialize_response

from starlette.background import BackgroundTasks
from starlette.concurrency import run_in_threadpool
from starlette.requests import Request
from starlette.responses import Response
from starlette.responses import JSONResponse as BaseJSONResponse
from starlette.routing import Match, request_response, compile_path

import aiojobs
import fastapi.params
import simdjson as json

logger = get_logger(name='LazyFastAPI')

_JSON_PARSER = json.Parser()
_JSON_DECODER = json.loads
_JSON_ENCODER = json.dumps

try:
    from functools import cached_property
except ImportError:
    class cached_property:  # noqa
        def __init__(self, func):
            self.__doc__ = getattr(func, "__doc__")
            self.func = func

        def __get__(self, obj, cls):
            if obj is None:
                return self
            value = obj.__dict__[self.func.__name__] = self.func(obj)
            return value

try:
    import sentry_sdk
    from sentry_sdk.utils import transaction_from_function as sentry_transaction_from_function
except ImportError:
    sentry_sdk = None
    sentry_transaction_from_function = None

def sjson_loads(data, *args, **kwargs):
    try:
        return _JSON_PARSER.parse(data, recursive=kwargs.get('recursive', False))
    except Exception as e:
        logger.error(f'Simdjson decoding error: {e}')
        return _JSON_DECODER.loads(data, *args, **kwargs)


def sjson_dumps(v, *, default):
    return json.dumps(v, default=default).decode()

class JSONResponse(BaseJSONResponse):
    def render(self, content: Any) -> bytes:
        return json.dumps(content, ensure_ascii=False, allow_nan=False, indent=None, separators=(",", ":")).encode("utf-8")

class BaseModel(PyBaseModel):
    class Config:
        json_loads = sjson_loads
        json_dumps = sjson_dumps


from functools import lru_cache, wraps
from typing import List, Optional, Dict
from asgiref.sync import async_to_sync
from .mp_utils import multiproc, _MAX_PROCS, lazy_parallelize


class LazyProc:
    def __init__(self, name, func, start=False, daemon=False, *args, **kwargs):
        self.name = name
        self.func = func
        self.start = start
        self.args = args
        self.kwargs = kwargs
        self.daemon = daemon
    
    @property
    def config(self):
        return {'target': self.func, 'daemon': self.daemon, 'args': self.args, 'kwargs': self.kwargs}


class LazyProcs:
    active = False
    active_procs: Dict[str, multiproc.Process] = {}
    inactive_procs: Dict[str, multiproc.Process] = {}

    @classmethod
    def set_state(cls):
        LazyProcs.active = bool(LazyProcs.num_active > 1)

    @classmethod
    def add_proc(cls, proc: LazyProc):
        process = multiproc.Process(**proc.config)
        if proc.start:
            process.start()
            LazyProcs.active_procs[proc.name] = process
        else:
            LazyProcs.inactive_procs[proc.name] = process
        LazyProcs.set_state()
        return process
    
    @classmethod
    def kill_proc(cls, name: str):
        if name in LazyProcs.active_procs:
            proc = LazyProcs.active_procs.pop(name)
            proc.terminate()
            LazyProcs.inactive_procs[name] = proc
            LazyProcs.set_state()
    
    @classmethod
    def killall(cls):
        for name in LazyProcs.active_procs:
            LazyProcs.kill_proc(name)
    
    @property
    def num_active(self):
        return len(self.active_procs)
    
    @property
    def num_inactive(self):
        return len(self.inactive_procs)

def lazyproc(name: str = 'lazyproc', start: bool = True, daemon: bool = False):
    def wrapper_proc(func):
        def wrapped_proc(*args, **kwargs):
            proc = LazyProc(name, start=start, daemon=daemon, *args, **kwargs)
            proc = LazyProcs.add_proc(proc)
            return proc
            #return func(*args, **kwargs)
        return wrapped_proc
    return wrapper_proc

def lazymultiproc(dataset = None, dataset_callable = None, dataset_var: str = None, num_procs: int = 10):
    def wrapper_multiproc(func):
        def wrapped_parallelize(*args, **kwargs):
            if dataset_callable:
                yield from lazy_parallelize(func, processes=num_procs, result=dataset_callable(*args, **kwargs), *args, **kwargs)
            else:
                datavar = dataset or kwargs.get(dataset_var)
                if datavar:
                    yield from lazy_parallelize(func, processes=num_procs, result=dataset, *args, **kwargs)
            return func(*args, **kwargs)
        return wrapped_parallelize
    return wrapper_multiproc



from ._base import *
from datetime import datetime
from typing import Optional, List, Union
from dataclasses import dataclass
from lazyops.lazyclasses.api import lazyclass
from lazyops.utils import timed_cache

_NP = None

# We lazily init news-please because it's a fat and heavy repo. Need to pick something else in the future.
@timed_cache(600)
def _init_newsplease():
    global _NP
    if _NP is not None:
        return
    nplease = lazy_init('news-please', 'newsplease')
    _NP = nplease.NewsPlease


@lazyclass
@dataclass
class Article:
    url: str
    title: str = None
    description: str = None
    image_url: str = None
    language: str = None
    domain: str = None
    text: str = None
    authors: List[str] = None
    date_publish: Union[str, datetime] = None
    date_modify: Union[str, datetime] = None
    extracted: Optional[bool] = False

    def _run(self):
        if self.extracted:
            return
        _init_newsplease()
        try:
            data = _NP.from_url(self.url)
            self.title = data.title
            self.description = data.description
            self.image_url = data.image_url
            self.language = data.language
            self.domain = data.source_domain
            self.text = data.maintext
            self.authors = data.authors
            self.date_publish = data.date_publish
            self.date_modify = data.date_modify
            self.extracted = True
        except Exception as e:
            logger.error(f'Error Parsing URL: {self.url}.\n{str(e)}')



@lazyclass
@dataclass
class GDELTArticle:
    url: str
    url_mobile: str = ''
    title: str = ''
    seendate: str = ''
    socialimage: str = ''
    domain: str = ''
    language: str = ''
    sourcecountry: str = ''
    text: Optional[str] = None
    extraction: Optional[Article] = None

    def parse(self):
        if self.extraction is not None:
            return
        self.extraction = Article(url=self.url)
        self.extraction._run()
        if self.extraction.extracted:
            self.text = self.extraction.text
    
    async def async_parse(self):
        if self.extraction is not None:
            return
        self.extraction = Article(url=self.url)
        self.extraction._run()
        if self.extraction.extracted:
            self.text = self.extraction.text



import re
from typing import Optional, List, Union, Optional, Any, Dict
from dataclasses import dataclass
from lazyops import get_logger, PathIO
from lazyops.lazyclasses import lazyclass


@lazyclass
@dataclass
class TFSModelVersion:
    step: int
    label: str = 'latest'


@lazyclass
@dataclass
class TFSModelConfig:
    name: str
    base_path: str
    model_platform: str = 'tensorflow'
    model_versions: Optional[List[TFSModelVersion]] = None

    def set_only_latest(self):
        if self.model_versions:
            self.model_versions = [version for version in self.model_versions if version.label == 'latest']
    
    @property
    def default_label(self):
        if self.model_versions:
            return self.model_versions[0].label
        return None

    @property
    def default_version(self):
        if self.model_versions:
            return str(self.model_versions[0].step)
        return None
    


from ._base import *


Filter = Union[List[str], str]


def near(n: int, *args) -> str:
    """
    Build the filter to find articles containing words that occur within
    `n` words of each other.
    eg. near(5, "airline", "climate") finds articles containing both
    "airline" and "climate" within 5 words.
    """
    if len(args) < 2:
        raise ValueError("At least two words must be provided")

    return f"near{str(n)}:" + '"' + " ".join(args) + '"'


def repeat(n: int, keyword: str) -> str:
    """
    Build the filter to find articles containing `keyword` at least `n` times.
    eg. repeat(2, "environment") finds articles containing the word "environment"
    at least twice.
    Only single word repetitions are allowed.
    """
    if " " in keyword:
        raise ValueError("Only single words can be repeated")

    return f'repeat{str(n)}:"{keyword}"'


def multi_repeat(repeats: List[Tuple[int, str]], method: str) -> str:
    """
    Build the filter to find articles containing multiple repeated words using `repeat()`
    eg. multi_repeat([(2, "airline"), (3, "airport")], "AND") finds articles that contain the word "airline" at least
    twice and "airport" at least 3 times.
    Params
    ------
        repeats: A list of (int, str) tuples to be passed to `repeat()`. Eg. [(2, "airline"), (3, "airport")]
        method: How to combine the restrictions. Must be one of "AND" or "OR"
    """
    if method not in ["AND", "OR"]:
        raise ValueError(f"method must be one of AND or OR, not {method}")

    to_repeat = [repeat(n, keyword) for (n, keyword) in repeats]
    return method.join(to_repeat)


class GDELTFilters:
    def __init__(
        self,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        timespan: Optional[str] = None,
        num_records: int = 250,
        keyword: Optional[Filter] = None,
        domain: Optional[Filter] = None,
        domain_exact: Optional[Filter] = None,
        near: Optional[str] = None,
        repeat: Optional[str] = None,
        country: Optional[Filter] = None,
        theme: Optional[Filter] = None,
    ) -> None:
        
        self.query_params: List[str] = []
        self._valid_countries: List[str] = []
        self._valid_themes: List[str] = []

        # Check we have either start/end date or timespan, but not both
        if not start_date and not end_date and not timespan:
            raise ValueError("Must provide either start_date and end_date, or timespan")

        if start_date and end_date and timespan:
            raise ValueError("Can only provide either start_date and end_date, or timespan")

        if keyword: self.query_params.append(self._keyword_to_string(keyword))

        if domain: self.query_params.append(self._filter_to_string("domain", domain))

        if domain_exact: self.query_params.append(self._filter_to_string("domainis", domain_exact))

        if country: self.query_params.append(self._filter_to_string("sourcecountry", country))

        if theme: self.query_params.append(self._filter_to_string("theme", theme))

        if near: self.query_params.append(near)

        if repeat: self.query_params.append(repeat)

        if start_date:
            self.query_params.append(f'&startdatetime={start_date.replace("-", "")}000000')
            self.query_params.append(f'&enddatetime={end_date.replace("-", "")}000000')
        else:
            # Use timespan
            self.query_params.append(f"&timespan={timespan}")

        if num_records > 250:
            raise ValueError(f"num_records must 250 or less, not {num_records}")

        self.query_params.append(f"&maxrecords={str(num_records)}")

    @property
    def query_string(self) -> str:
        return "".join(self.query_params)

    @staticmethod
    def _filter_to_string(name: str, f: Filter) -> str:
        if type(f) == str:
            return f"{name}:{f}"

        else:
            # Build an OR statement
            return "(" + " OR ".join(f"{name}:{clause}" for clause in f) + ")"

    @staticmethod
    def _keyword_to_string(keywords: Filter) -> str:
        if type(keywords) == str:
            return f'"{keywords}"'

        else:
            return (
                "("
                + " OR ".join(
                    f'"{word}"' if " " in word else word for word in keywords
                )
            ) + ")"




import sys
import types
import typing
import warnings
import simdjson as json

from enum import Enum
from dataclasses import is_dataclass


from .utils import issubclass_safe

_FROM = 1
_TO = 2
_PARSER = json.Parser()

def _process_class(cls):
    _process_class_internal(cls)
    cls.from_dict = cls._lazyclasses_from_dict
    cls.to_dict = cls._lazyclasses_to_dict

    def from_json(cls, json_data):
        return cls.from_dict(_PARSER.parse(json_data))

    def to_json(self, *, separators=None, indent=None):
        if indent is None and separators is None: separators = (',', ':')
        return json.dumps(self.to_dict(), separators=separators, indent=indent)

    cls.from_json = classmethod(from_json)
    cls.to_json = to_json
    return cls


def _process_class_internal(cls):
    # Delay the building of our from_dict method until it is first called.
    # This allows the compilation to reference classes defined later in
    # the module.
    def _temp_from_dict(cls, *args, **kwarg):
        _replace_from_dict(cls, '_lazyclasses_from_dict')
        return cls._lazyclasses_from_dict(*args, **kwarg)

    cls.from_dict = classmethod(_temp_from_dict)
    cls._lazyclasses_from_dict = classmethod(_temp_from_dict)

    def _temp_to_dict(self, *args, **kwargs):
        _replace_to_dict(cls, '_lazyclasses_to_dict')
        return self._lazyclasses_to_dict(*args, **kwargs)

    cls.to_dict = _temp_to_dict
    cls._lazyclasses_to_dict = _temp_to_dict

    return cls


def _replace_from_dict(cls, from_dict='from_dict'):
    from_dict_src = _from_dict_source(cls)
    from_dict_module = compile(from_dict_src, '<lazyclass_generated_code>', 'exec')
    from_dict_code = [const for const in from_dict_module.co_consts if isinstance(const, types.CodeType)][0]
    the_globals = {
        **sys.modules[cls.__module__].__dict__,
        **referenced_types(cls)
    }

    from_dict_func = types.FunctionType(from_dict_code, the_globals, from_dict)
    setattr(cls, from_dict, classmethod(from_dict_func))


def _replace_to_dict(cls, to_dict='to_dict'):
    to_dict_src = _to_dict_source(cls)
    to_dict_module = compile(to_dict_src, '<lazyclass_generated_code>', 'exec')
    to_dict_code = [const for const in to_dict_module.co_consts if isinstance(const, types.CodeType)][0]
    to_dict_func = types.FunctionType(to_dict_code, sys.modules[cls.__module__].__dict__, to_dict,)

    setattr(cls, to_dict, to_dict_func)


def _from_dict_source(cls):
    lines = [
        'def from_dict(cls, o):',
        '    args = []',
    ]
    for name, field_type in typing.get_type_hints(cls).items():
        # pop off the top layer of optional, since we are using o.get
        if typing.get_origin(field_type) == typing.Union: field_type = typing.get_args(field_type)[0]

        access = f'o.get({name!r})'
        transform = expr_builder_from(field_type)

        if transform('x') != 'x':
            lines.append(f'    value = {access}')
            lines.append(f'    if value is not None:')  # noqa: F541
            lines.append(f'        value = ' + transform('value'))  # noqa: E501,F541
            lines.append(f'    args.append(value)')  # noqa: F541
        else:
            lines.append(f'    args.append({access})')
    
    lines.append('    return cls(*args)')
    lines.append('')
    return '\n'.join(lines)


def _to_dict_source(cls):
    lines = [
        'def to_dict(self):',
        '    result = {}',
    ]
    INCLUDE_NONES = False

    for name, field_type in typing.get_type_hints(cls).items():
        access = f'self.{name}'
        transform = expr_builder_to(field_type)

        if transform('x') != 'x':
            # since we have an is not none check, elide the first level
            # of optional
            if typing.get_origin(field_type) == typing.Union: transform = expr_builder_to(typing.get_args(field_type)[0])
            
            lines.append(f'    value = {access}')
            lines.append(f'    if value is not None:')  # noqa: F541
            lines.append(f'        value = ' + transform('value'))  # noqa: E501,F541
            
            if INCLUDE_NONES:
                lines.append(f'    result[{name!r}] = value')
            
            else:
                lines.append(f'        result[{name!r}] = value')
        else:
            lines.append(f'    result[{name!r}] = {access}')

    lines.append('    return result')
    lines.append('')
    return '\n'.join(lines)


def expr_builder_from(t: type, depth=0):
    return expr_builder(t, depth, direction=_FROM)


def expr_builder_to(t: type, depth=0):
    return expr_builder(t, depth, direction=_TO)


def expr_builder(t: type, depth=0, direction=_FROM):
    def identity(expr):
        return expr

    origin = typing.get_origin(t)

    if origin == typing.Union:
        type_arg = typing.get_args(t)[0]
        inner = expr_builder(type_arg, depth + 1, direction)

        def f(expr):
            t0 = f'__{depth}'
            return f'{inner(t0)} if ({t0}:=({expr})) is not None else None'

        return f
    
    elif origin == list:
        type_arg = typing.get_args(t)[0]
        inner = expr_builder(type_arg, depth + 1, direction)

        def f(expr):
            t0 = f'__{depth}'
            return f'[{inner(t0)} for {t0} in {expr}]'
        return f
    
    elif origin == dict:
        key_type, value_type = typing.get_args(t)

        if not issubclass_safe(key_type, str):
            warnings.warn(f'to_json will not work for non-str key dict: {t}')
            return identity

        inner = expr_builder(value_type, depth + 1, direction)

        def f(expr):
            k0 = f'__k{depth}'
            v0 = f'__v{depth}'
            return (
                '{'
                + f'{k0}: {inner(v0)} for {k0},{v0} in ({expr}).items()'
                + '}'
            )
        return f
    
    elif is_dataclass(t):
        # Give indirectly referenced dataclassses a to_dict method without
        # trashing their public API
        if not hasattr(t, '_lazyclasses_from_dict'):
            _process_class_internal(t)

        # TODO: consider calling to_dict, from_dict if they are there
        # i.e. create a way for serialization to be overridden

        if direction == _FROM:
            def f(expr):
                return f'{t.__name__}._lazyclasses_from_dict({expr})'
            return f
        else:
            def f(expr):
                # or should be have a function that takes the class and its
                # type?
                return f'({expr})._lazyclasses_to_dict()'
            return f
    elif issubclass_safe(t, Enum):
        if direction == _FROM:
            def f(expr):
                return f'{t.__name__}({expr})'
            return f
        else:
            def f(expr):
                return f'({expr}).value'
            return f

    from datetime import date, datetime
    if issubclass_safe(t, datetime):
        if direction == _FROM:
            def f(expr):
                t0 = f'__{depth}'
                return (f'{t.__name__}.fromisoformat('
                        f'{t0}[:-1]+"+00:00" if ({t0}:={expr})[-1]=="Z" '
                        f'else {t0}'
                        ')')
            return f
        else:
            return lambda expr: f'({expr}).isoformat()'
    if issubclass_safe(t, date):
        if direction == _FROM:
            # TODO: use dateutil.parser.isoparse if available?
            return lambda expr: f'{t.__name__}.fromisoformat({expr})'
        else:
            return lambda expr: f'({expr}).isoformat()'

    return identity


def referenced_types(cls):
    from datetime import date, datetime

    def extract_type(t):
        origin = typing.get_origin(t)
        if origin in [typing.Union, list]:
            type_arg = typing.get_args(t)[0]
            return extract_type(type_arg)

        elif origin == dict:
            value_type_arg = typing.get_args(t)[1]
            return extract_type(value_type_arg)

        elif is_dataclass(t) or issubclass_safe(t, (Enum, date, datetime)):
            return t

        return None

    types = {}
    for _, field_type in typing.get_type_hints(cls).items():
        t = extract_type(field_type)
        if t is not None: types[t.__name__] = t
    return types



from ._base import *
from .models import Params


def fix_query_dependencies(dependant: Dependant):
    dependant.body_params.extend(dependant.query_params)
    dependant.query_params = []

    for field in dependant.body_params:
        if not isinstance(field.field_info, Params):
            field.field_info.embed = True

    for sub_dependant in dependant.dependencies:
        fix_query_dependencies(sub_dependant)


def clone_dependant(dependant: Dependant) -> Dependant:
    new_dependant = Dependant()
    new_dependant.path_params = dependant.path_params
    new_dependant.query_params = dependant.query_params
    new_dependant.header_params = dependant.header_params
    new_dependant.cookie_params = dependant.cookie_params
    new_dependant.body_params = dependant.body_params
    new_dependant.dependencies = dependant.dependencies
    new_dependant.security_requirements = dependant.security_requirements
    new_dependant.request_param_name = dependant.request_param_name
    new_dependant.websocket_param_name = dependant.websocket_param_name
    new_dependant.response_param_name = dependant.response_param_name
    new_dependant.background_tasks_param_name = dependant.background_tasks_param_name
    new_dependant.security_scopes = dependant.security_scopes
    new_dependant.security_scopes_param_name = dependant.security_scopes_param_name
    new_dependant.name = dependant.name
    new_dependant.call = dependant.call
    new_dependant.use_cache = dependant.use_cache
    new_dependant.path = dependant.path
    new_dependant.cache_key = dependant.cache_key
    return new_dependant


def insert_dependencies(target: Dependant, dependencies: Sequence[Depends] = None):
    assert target.path
    if not dependencies:
        return
    for depends in dependencies[::-1]:
        target.dependencies.insert(
            0,
            get_parameterless_sub_dependant(depends=depends, path=target.path),
        )



from ._base import *
from .models import LazyDBCacheBase

logger = get_logger('LazyDB', module='backends')

class PklDBCache(LazyDBCacheBase):
    def dumps(self, data, *args, **kwargs):
        return pkler.dumps(data, *args, **kwargs)
    
    def loads(self, data, *args, **kwargs):
        return pkler.loads(data, *args, **kwargs)

    def restore(self, *args, **kwargs):
        self.lock.acquire()
        dbdata = {}
        with self.lock:
            try:
                dbdata = fio.pklload(self.cache_filepath)
            except Exception as e:
                logger.error(f'Failed to Restore DB {self._cachename}: {str(e)}')
                logger.error(f'Copying DB to Backup')
                tempfn = fio.mod_fname(self.cache_filepath, suffix='_' + tstamp())
                fio.copy(self.cache_filepath, tempfn)
                logger.error(f'DB saved to {tempfn}')
        self.lock.release()
        return dbdata
    
    def save(self, dbdata, *args, **kwargs):
        _saved = False
        self.lock.acquire()
        with self.lock:
            try:
                fio.pklsave(dbdata, self.cache_filepath)
                _saved = True
            except Exception as e:
                logger.error(f'Failed to Save DB {self._cachename}: {str(e)}')
                _saved = False
        self.lock.release()
        return _saved

class GSheetsDBCache(LazyDBCacheBase):
    def __init__(self, sheets_url=None, cache_name='LazyCacheDB', auth_path=None, *args, **kwargs):
        self.url = sheets_url
        self.auth = auth_path
        self.cache_name = cache_name
        self.indexes = {}
        self.setup_client()

    def setup_client(self):
        gspread = lazy_init('gspread')
        self.gc = gspread.service_account(filename=self.auth)
        if self.url:
            self.sheet = self.gc.open_by_url(self.url)
            logger.info(f'Loaded GSheetsDB from URL: {self.url}')
        else:
            try:
                self.sheet = self.gc.open(self.cache_name)
                logger.info(f'Loaded GSheetsDB from Name: {self.cache_name}')
            
            except Exception as e:
                self.sheet = None
                logger.error(f'Failed to Load GSheetsDB from Name: {self.cache_name}')
        
        if not self.sheet:
            self.sheet = self.gc.create(self.cache_name)
            logger.info(f'Created GSheetsDB by Name: {self.cache_name}')
            self.url = self.sheet.url
        self.refresh_index()
    
    def create_indexes(self, dbdata):
        for n, i in enumerate(dbdata):
            if i not in self.all_wks:
                idx = dbdata[i]
                wks = self.sheet.add_worksheet(i, index=n)
                # Create row 0 - header
                wks.append_row(idx.schema_props)

    def dump_indexes(self, dbdata):
        for schema, idx in dbdata.items():
            wks = self.sheet.worksheet(schema)
            items = []
            for i in idx.index.values():
                d = i.dict()
                items.append(list(d.values()))
            wks.insert_rows(items, row=2)
        self.refresh_index()

    def refresh_index(self):
        self.index = self.all_wks_dict
    
    def get_header(self, wks):
        return wks.row_values(1)

    @property
    def all_wks(self):
        return list(self.sheet.worksheets())
    
    @property
    def all_wks_dict(self):
        return {n: s for n,s in enumerate(self.all_wks)}
    
    @property
    def cache_file(self):
        return None

    @property
    def cache_filepath(self):
        return None
    
    @property
    def exists(self):
        return bool(self.url)
    
    # Need to figure out how to reconstruct the data
    def dumps(self, data, *args, **kwargs):
        pass
    
    def loads(self, *args, **kwargs):
        pass

    def restore(self, *args, **kwargs):
        pass
    
    def save(self, db, *args, **kwargs):
        pass



from ._base import *
from .models import component_name, rename_if_scope_child_component
from fastapi import status

class BaseError(Exception):
    CODE = None
    MESSAGE = None

    ErrorModel = None
    DataModel = None

    data_required = False
    errors_required = False

    error_model = None
    data_model = None
    resp_model = None

    _component_name = None

    def __init__(self, data=None):
        if data is None: data = {}
        raw_data = data
        data = self.validate_data(raw_data)
        Exception.__init__(self, self.CODE, self.MESSAGE)
        self.data = data
        self.raw_data = raw_data

    @classmethod
    def validate_data(cls, data):
        data_model = cls.get_data_model()
        if data_model: data = data_model.validate(data)
        return data

    def __str__(self):
        s = f"[{self.CODE}] {self.MESSAGE}"
        if self.data: s += f": {self.data!r}"
        return s

    def get_resp_data(self):
        return self.raw_data

    @classmethod
    def get_description(cls):
        s = cls.get_default_description()
        if cls.__doc__: s += "\n\n" + cls.__doc__
        return s

    @classmethod
    def get_default_description(cls):
        return f"[{cls.CODE}] {cls.MESSAGE}"

    def get_resp(self):
        error = {
            'code': self.CODE,
            'message': self.MESSAGE,
        }

        resp_data = self.get_resp_data()
        if resp_data: error['data'] = resp_data

        resp = {
            'jsonrpc': '2.0',
            'error': error,
            'id': None,
        }

        return jsonable_encoder(resp)

    @classmethod
    def get_error_model(cls):
        if cls.__dict__.get('error_model') is not None:
            return cls.error_model
        cls.error_model = cls.build_error_model()
        return cls.error_model

    @classmethod
    def build_error_model(cls):
        if cls.ErrorModel is not None:
            return rename_if_scope_child_component(cls, cls.ErrorModel, 'Error')
        return None

    @classmethod
    def get_data_model(cls):
        if cls.__dict__.get('data_model') is not None:
            return cls.data_model
        cls.data_model = cls.build_data_model()
        return cls.data_model

    @classmethod
    def build_data_model(cls):
        if cls.DataModel is not None:
            return rename_if_scope_child_component(cls, cls.DataModel, 'Data')

        error_model = cls.get_error_model()
        if error_model is None:
            return None

        errors_annotation = List[error_model]
        if not cls.errors_required: errors_annotation = Optional[errors_annotation]

        ns = {
            '__annotations__': {
                'errors': errors_annotation,
            }
        }

        _ErrorData = ModelMetaclass.__new__(ModelMetaclass, '_ErrorData', (BaseModel,), ns)
        _ErrorData = component_name(f'_ErrorData[{error_model.__name__}]', error_model.__module__)(_ErrorData)

        return _ErrorData

    @classmethod
    def get_resp_model(cls):
        if cls.__dict__.get('resp_model') is not None:
            return cls.resp_model
        cls.resp_model = cls.build_resp_model()
        return cls.resp_model

    @classmethod
    def build_resp_model(cls):
        ns = {
            'code': Field(cls.CODE, const=True, example=cls.CODE),
            'message': Field(cls.MESSAGE, const=True, example=cls.MESSAGE),
            '__annotations__': {
                'code': int,
                'message': str,
            }
        }

        data_model = cls.get_data_model()
        if data_model is not None:
            if not cls.data_required: data_model = Optional[data_model]
            # noinspection PyTypeChecker
            ns['__annotations__']['data'] = data_model

        name = cls._component_name or cls.__name__

        _JsonRpcErrorModel = ModelMetaclass.__new__(ModelMetaclass, '_JsonRpcErrorModel', (BaseModel,), ns)
        _JsonRpcErrorModel = component_name(name, cls.__module__)(_JsonRpcErrorModel)

        @component_name(f'_ErrorResponse[{name}]', cls.__module__)
        class _ErrorResponseModel(BaseModel):
            jsonrpc: StrictStr = Field('2.0', const=True, example='2.0')
            id: Union[StrictStr, int] = Field(None, example=0)
            error: _JsonRpcErrorModel
            class Config:
                extra = 'forbid'

        return _ErrorResponseModel

@component_name('_Error')
class ErrorModel(BaseModel):
    loc: List[str]
    msg: str
    type: str
    ctx: Optional[Dict[str, Any]]


class ParseError(BaseError):
    """Invalid JSON was received by the server"""
    CODE = status.HTTP_422_UNPROCESSABLE_ENTITY
    #CODE = -32700
    MESSAGE = "Parse error"


class InvalidRequest(BaseError):
    """The JSON sent is not a valid Request object"""
    #CODE = -32600
    CODE = status.HTTP_400_BAD_REQUEST
    MESSAGE = "Invalid Request"
    error_model = ErrorModel


class MethodNotFound(BaseError):
    """The method does not exist / is not available"""
    #CODE = -32601
    CODE = status.HTTP_405_METHOD_NOT_ALLOWED
    MESSAGE = "Method not found"


class InvalidParams(BaseError):
    """Invalid method parameter(s)"""
    #CODE = -32602
    CODE = status.HTTP_424_FAILED_DEPENDENCY
    MESSAGE = "Invalid params"
    error_model = ErrorModel


class InternalError(BaseError):
    """Internal JSON-RPC error"""
    #CODE = -32603
    CODE = status.HTTP_500_INTERNAL_SERVER_ERROR
    MESSAGE = "Internal error"


class NoContent(Exception):
    pass

def errors_responses(errors: Sequence[Type[BaseError]] = None):
    responses = {'default': {}}
    if errors:
        cnt = 1
        for error_cls in errors:
            #responses[404] = {
            #    'model': error_cls.get_resp_model(),
            #    'description': error_cls.get_description(),
            #}
            responses[error_cls.CODE] = {
                'model': error_cls.get_resp_model(),
                'description': error_cls.get_description(),
            }
            
            #responses[f'200{" " * cnt}'] = {
            #    'model': error_cls.get_resp_model(),
            #    'description': error_cls.get_description(),
            #}
            cnt += 1
    return responses


def invalid_request_from_validation_error(exc: ValidationError) -> InvalidRequest:
    return InvalidRequest(data={'errors': exc.errors()})


def invalid_params_from_validation_error(exc: ValidationError) -> InvalidParams:
    errors = []

    for err in exc.errors():
        if err['loc'][:1] == ('body', ):
            err['loc'] = err['loc'][1:]
        else:
            assert err['loc']
            err['loc'] = (f"<{err['loc'][0]}>", ) + err['loc'][1:]
        errors.append(err)

    return InvalidParams(data={'errors': errors})





from lazyops import lazy_init, get_logger, timed_cache, LazyObject
lazy_init('pandas')

import pandas as pd

from enum import Enum
from typing import Dict, Optional, List, Union, Tuple

from dataclasses import dataclass
from lazyops.apis import LazySession, async_req
from lazyops.lazyio import LazyJson
from lazyops.lazyclasses import lazyclass


logger = get_logger('LazySources', 'GDELT')




import os
import sys
import asyncio
import threading
from uuid import uuid4
from datetime import datetime, timedelta
from functools import lru_cache, wraps
from lazyops.models import LazyData
from lazyops.common import lazy_import, lazylibs
from lazyops.retry import retryable

def create_uuid(name=None, sep='-', strip=False):
    uid = str(uuid4())
    if name: uid = name + sep + uid
    if strip: uid.replace(sep, '').strip().replace('-', '').strip()
    return uid

def latest_tf_ckpt(model_path):
    ckpt_mtg = lazy_import('tensorflow.python.training.checkpoint_management')
    latest = ckpt_mtg.latest_checkpoint(model_path)
    ckpt_num = int(latest.split('/')[-1].split('-')[-1])
    return LazyData(string=f"Latest Checkpoint = Step {ckpt_num} @ {latest}", value={'step': ckpt_num, 'path': model_path, 'latest_ckpt': latest}, dtype="tfcheckpoint")

def build_chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def list_to_dict(items, delim='='):
    res = {}
    for item in items:
        i = item.split(delim, 1)
        res[i[0].strip()] = i[-1].strip()
    return res


def get_variable_separator():
    """
    Returns the environment variable separator for the current platform.
    :return: Environment variable separator
    """
    if sys.platform.startswith('win'):
        return ';'
    return ':'


def find_binary_in_path(filename):
    """
    Searches for a binary named `filename` in the current PATH. If an executable is found, its absolute path is returned
    else None.
    :param filename: Filename of the binary
    :return: Absolute path or None
    """
    if 'PATH' not in os.environ:
        return None
    for directory in os.environ['PATH'].split(get_variable_separator()):
        binary = os.path.abspath(os.path.join(directory, filename))
        if os.path.isfile(binary) and os.access(binary, os.X_OK):
            return binary
    return None



def timed_cache(seconds: int, maxsize: int = 128):
    def wrapper_cache(func):
        func = lru_cache(maxsize=maxsize)(func)
        func.lifetime = timedelta(seconds=seconds)
        func.expiration = datetime.utcnow() + func.lifetime
        @wraps(func)
        def wrapped_func(*args, **kwargs):
            if datetime.utcnow() >= func.expiration:
                func.cache_clear()
                func.expiration = datetime.utcnow() + func.lifetime
            return func(*args, **kwargs)
        return wrapped_func
    return wrapper_cache


def require_module(name, *args, **kwargs):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            submod = lazylibs.get_submodule(name)
            if not submod:
                submod = lazylibs.setup_submodule(name, *args, **kwargs)
            if not submod.has_initialized:
                submod.lazy_init()            
            return f(*args, **kwargs)
        return wrapper
    return decorator



from ._base import *

# make stuff compatible and compliant with posix
# https://www.tensorflow.org/datasets/api_docs/python/tfds/core/ReadWritePath?hl=en

class LazyFile:
    def __init__(self, filename, method='tmp', overwrite=True, cleanup=True):
        self._dest_filename = filename
        self._method = method
        self.overwrite = overwrite
        self.cleanup = cleanup
        self._tmp_file = NamedTemporaryFile(delete=False) if method == 'tmp' else BytesIO()
    
    @property
    def tmpfile(self):
        return self._tmp_file
    
    @property
    def tmpfile_path(self):
        if self._tmp_file and self._method == 'tmp':
            return self._tmp_file.name
        return None
    
    @property
    def is_closed(self):
        return bool(self._tmp_file is not None)

    def write(self):
        assert not self.is_closed
        assert File.exists(self._dest_filename) is False or self.overwrite is True
        if self._method == 'tmp':
            data = gfile(self._tmp_file.name, 'rb').read()
        else:
            data = self._tmp_file.getvalue()
        with gfile(self._dest_filename, 'wb') as f:
            f.write(data)
        if self.cleanup and self._method == 'tmp':
            os.remove(self._tmp_file.name)
        self._tmp_file = None



class LazyJson:
    serializer = json
    parser = json.Parser()

    @classmethod
    def dumps(cls, obj, *args, **kwargs):
        return LazyJson.serializer.dumps(obj, *args, **kwargs)
    
    @classmethod
    def dump(cls, obj, fileio, indent=2, *args, **kwargs):
        return LazyJson.serializer.dump(obj, fp=fileio, indent=indent, *args, **kwargs)

    @classmethod
    def loads(cls, jsonstring, use_parser=True, recursive=False, ignore_errors=True, *args, **kwargs):
        if use_parser:
            try:
                return LazyJson.parser.parse(jsonstring, recursive=recursive)
            except Exception as e:
                if not ignore_errors:
                    raise ValueError(str(e))
        try:
            return LazyJson.serializer.loads(jsonstring, *args, **kwargs)
        except Exception as e:
            if not ignore_errors:
                raise ValueError(str(e))
        return None

    @classmethod
    def load(cls, fileio, use_parser=True, recursive=False, ignore_errors=True, *args, **kwargs):
        if use_parser:
            try:
                return LazyJson.parser.parse(fileio, recursive=recursive)
            except Exception as e:
                if not ignore_errors:
                    raise ValueError(str(e))
        try:
            return LazyJson.serializer.load(fileio, *args, **kwargs)
        except Exception as e:
            if not ignore_errors:
                raise ValueError(str(e))
        return None

class LazyPickler:
    serializer = pickler
    protocol = pickler.HIGHEST_PROTOCOL

    @classmethod
    def dumps(cls, obj, *args, **kwargs):
        return LazyPickler.serializer.dumps(obj, protocol=LazyPickler.protocol, *args, **kwargs)
    
    @classmethod
    def loads(cls, data, *args, **kwargs):
        return LazyPickler.serializer.loads(data, *args, **kwargs)
    
    @classmethod
    def dump(cls, obj, fileio, *args, **kwargs):
        data = LazyPickler.dumps(obj,  *args, **kwargs)
        fileio.write(data)
        fileio.flush()

    @classmethod
    def load(cls, fileio, *args, **kwargs):
        return LazyPickler.loads(fileio.read(), *args, **kwargs)



class ModeIO(Enum):
    r = 'r'
    rb = 'rb'
    read = 'r'
    read_binary = 'rb'
    a = 'a'
    append = 'a'
    w = 'w'
    wb = 'wb'
    write = 'w'
    write_binary = 'wb'
    rw = 'r+'
    rwb = 'rb+'
    readwrite = 'r+'
    readwrite_binary = 'rb+'
    auto = 'auto'

class FileExtIO(Enum):
    TXT = ['txt', 'text']
    JSON = ['json']
    JSONLINES = ['jsonl', 'jsonlines', 'jlines']
    TORCH = ['bin', '.model']
    PICKLE = ['pkl', 'pickle', 'pb']
    NUMPY = ['numpy', 'npy']
    TFRECORDS = ['tfrecord', 'tfrecords', 'tfr']


class ExtIO:
    def __init__(self, ext: str):
        self._ext = None
        self._src_ext = ext
        for e in FileExtIO:
            if ext in e.value:
                self._ext = FileExtIO[e.name]
                break

    @property
    def appendable_fs(self):
        return [FileExtIO.TXT, FileExtIO.JSONLINES, FileExtIO.TFRECORDS]
    
    @property
    def appendable(self):
        return self._ext in self.appendable_fs 



class LazyIOBase(object):
    __metaclass__ = ABCMeta

    def __init__(
        self,
        filename: str,
        mode: ModeIO = ModeIO.auto,
        allow_deletion: bool = False,
        *args, **kwargs):
        self._setup_file(filename)
        self._mode = ModeIO.append if mode == 'auto' and self.exists else mode
        
        self._io = None
        self._io_closed = True
        self._allow_rm = allow_deletion

        self._args, self._kwargs = args, kwargs
        self.open()
    

    def _setup_file(self, filename: str = None):
        if filename is None:
            return
        self._filename = filename
        self._basename = File.base(filename)
        self._directory = File.getdir(filename)
        File.mkdirs(self._directory)
        self._fext = File.ext(self._filename)
        self._ext = ExtIO(self._fext)


    def open(self, filename: str = None, mode: ModeIO = None):
        mode = mode.value or self.mode
        filename = filename or self._filename
        if not self.is_closed:
            self.close()
        self._io = gfile(self._filename, mode)
        self._io_closed = False
        if not self.exists:
            self._io.write()
            self.flush()

    def close(self):
        if self._io is None:
            return
        self._io.flush()
        self._io.close()
        self._io_closed = True
        self._io = None
    
    def flush(self):
        self._io.flush()
    
    # actual write method per class
    def _write(self, data, *args, **kwargs):
        self._io.write(data)

    def write(self, data, *args, **kwargs):
        self._ensure_open()
        self._write(data, *args, **kwargs)

    def _read(self, *args, **kwargs):
        return self._io.read(*args, **kwargs)
    
    def read(self, *args, **kwargs):
        self._ensure_open()
        return self._read(*args, **kwargs)

    def _readlines(self, *args, **kwargs):
        return self._io.readlines()

    def readlines(self,  *args, **kwargs):
        self._ensure_open()
        return self._readlines(*args, **kwargs)
    
    @timed_cache(120)
    def get_num_lines(self):
        return sum(1 for _ in File.tflines(self._filename))

    @property
    def filesize(self):
        self._ensure_open()
        return self._io.size()

    @timed_cache(10)
    def seek(self, offset=None, whence=0, position=None):
        self._ensure_open()
        return self._io.seek(offset, whence, position)

    def _readline(self, *args, **kwargs):
        return self._io.readline()

    def readline(self, *args, **kwargs):
        self._ensure_open()
        return self._readline(*args, **kwargs)

    def _readlines(self, *args, **kwargs):
        return self._io.readline()

    def readlines(self, *args, **kwargs):
        self._ensure_open()
        return self._readlines(*args, **kwargs)
    
    def _tell(self, *args, **kwargs):
        return self._io.tell()

    def tell(self, *args, **kwargs):
        self._ensure_open()
        return self._tell(*args, **kwargs)
    
    def _iterator(self):
        return self

    def __iter__(self):
        return self._iterator()
    
    def _getnext(self):
        return self._io.next()

    def __next__(self):
        return self._getnext()
    
    @property
    def seekable(self):
        self._ensure_open()
        return self._io.seekable()


    def setmode(self, mode: ModeIO = ModeIO.auto):
        if mode.value != self._mode:
            self.close()
        self._mode = mode.value
        self.open()
    
    def setfile(self, filename: str, mode: ModeIO = None, allow_deletion: bool = False, *args, **kwargs):
        self.close()
        self._setup_file(filename)
        if mode is not None:
            self.setmode(mode)
        self._allow_rm = self._allow_rm or allow_deletion
        self._args = args
        if kwargs: self._kwargs.update(kwargs)
        self.open()

    @classmethod
    def modify(cls, filename, new_filename=None, prefix=None, suffix=None, extension=None, directory=None, create_dirs=True, filename_only=False, space_replace=None):
        return File.mod_fname(filename=filename, newname=new_filename, prefix=prefix, suffix=suffix, ext=extension, directory=directory, create_dirs=create_dirs, filename_only=filename_only, space_replace=space_replace)

    # ensures file is still open to do op. Caches clear every 15 secs.
    @timed_cache(15)
    def _ensure_open(self):
        if self.is_closed:
            raise ValueError('File is closed.')

    def compare(self, target_filename):
        from tensorflow.python.lib.io.file_io import filecmp
        return filecmp(self._filename, target_filename)

    @property
    def exists(self):
        return File.exists(self._filename)
    
    @property
    def mode(self):
        return self._mode.value
    
    @property
    def is_closed(self):
        return self._io_closed
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        self.backend = None

    def __del__(self):
        self.close()
        if not self._allow_rm:
            raise ValueError('Config allow_deletion = True must be set to allow del')
        File.rm(self._filename)
    
    def backup(self, filepath: str = None, directory: str = None, suffix: str = 'timestamp', overwrite: bool = False):
        if suffix == 'timestamp': suffix = tstamp()
        backup_fname = self.modify(self._filename, filepath=filepath, directory=directory, suffix=suffix)
        try:
            File.copy(self._filename, backup_fname, overwrite=overwrite)
        except Exception as e:
            logger.error(str(e))
            return None
        return backup_fname


class LazyIOText(LazyIOBase):
    newline = '\n'
    num_write_flush = 1000 # write ops before flushing
    
    def __init__(
        self, 
        filename: str,
        mode: ModeIO = ModeIO.auto,
        allow_deletion: bool = False,
        newline = None,
        num_write_flush = None,
        *args, **kwargs):

        self._io_writes = 0
        self.newline = newline or LazyIOText.newline
        self._num_write_flush = num_write_flush or LazyIOText.num_write_flush
        super(LazyIOText, self).__init__(filename, mode, allow_deletion, *args, **kwargs)
    
    def close(self):
        super(LazyIOText, self).close()
        self._io_writes = 0
    
    def _write(self, data, newline=None, flush=False, *args, **kwargs):
        newline = newline or self.newline
        self._io.write(data)
        self._io.write(newline)
        self._io_writes += 1
        if flush or (self._io_writes % self._num_write_flush == 0):
            self.flush()
    
    def _readlines(self, as_list=True, strip_newline=True, remove_empty_lines=True, *args, **kwargs):
        if not as_list:
            return self._read()
        texts = self._io.readlines()
        if strip_newline:
            texts = [t.strip() for t in texts]
        if remove_empty_lines:
            texts = [t for t in texts if t]
        return texts

class LazyIOJson(LazyIOBase):
    serializer = LazyJson
    
    def __init__(
        self, 
        filename: str,
        mode: ModeIO = ModeIO.read,
        allow_deletion: bool = False,
        serializer = None,
        *args, **kwargs):
        self.serializer = serializer or LazyIOJson.serializer
        super(LazyIOJson, self).__init__(filename, mode, allow_deletion, *args, **kwargs)
    
    def _write(self, data, dump_kwargs={}, *args, **kwargs):
        if self.mode.value not in ['w', 'wb']:
            self.setmode(mode=(ModeIO.write_binary if 'b' in self.mode.value else ModeIO.write))
        self.serializer.dump(data, self._io, **dump_kwargs)
        self.flush()
    
    def _read(self, load_kwargs={}, *args, **kwargs):
        return self.serializer.load(self._io, **load_kwargs)


class LazyIOJsonLines(LazyIOBase):
    newline = '\n'
    num_write_flush = 1000 # write ops before flushing
    serializer = LazyJson

    def __init__(
        self, 
        filename: str,
        mode: ModeIO = ModeIO.auto,
        allow_deletion: bool = False,
        newline = None,
        num_write_flush = None,
        serializer = None,
        *args, **kwargs):

        self._io_writes = 0
        self._line_map = []
        self.newline = newline or LazyIOJsonLines.newline
        self._num_write_flush = num_write_flush or LazyIOJsonLines.num_write_flush
        self.serializer = serializer or LazyIOJsonLines.serializer
        super(LazyIOJsonLines, self).__init__(filename, mode, allow_deletion, *args, **kwargs)

    def close(self):
        super(LazyIOJsonLines, self).close()
        self._io_writes = 0
    
    def _build_line_mapping(self):
        self._line_map.append(0)
        while self._io.readline():
            self._line_map.append(self._io.tell())
    
    def _write(self, data, newline=None, flush=False, dumps_kwargs={}, *args, **kwargs):
        newline = newline or self.newline
        self._io.write(self.serializer.dumps(data, **dumps_kwargs))
        self._io.write(newline)
        self._io_writes += 1
        if flush or (self._io_writes % self._num_write_flush == 0):
            self.flush()
    
    def _iterator(self, ignore_errors=True, loads_kwargs={}, *args, **kwargs):
        for line in self._io:
            try:
                line = self.serializer.loads(line, ignore_errors=ignore_errors, **loads_kwargs)
                if line: yield line
            except StopIteration:
                break
            except Exception as e:
                if not ignore_errors:
                    raise ValueError(str(e))
        

    def _readlines(self, as_iter=False, as_list=True, ignore_errors=True, loads_kwargs={}, *args, **kwargs):
        if as_iter:
            return self._iterator(ignore_errors=ignore_errors, loads_kwargs=loads_kwargs)
        if as_list:
            return [i for i in self._iterator(ignore_errors=ignore_errors, loads_kwargs={'recursive': True})]
        return self._io.readlines()

    def __len__(self):
        if not self._line_map:
            self._build_line_mapping()
        return len(self._line_map)

    @timed_cache(120)
    def get_num_lines(self):
        return self.__len__

    @timed_cache(120)
    def __getitem__(self, idx):
        if not self._line_map:
            self._build_line_mapping()
        self._f.seek(self._line_map[idx])
        return self.serializer.loads(self._f.readline(), ignore_errors=True, recursive=True)


class LazyIOPickle(LazyIOBase):
    
    serializer = LazyPickler
    num_write_flush = 1000 # write ops before flushing
    
    def __init__(
        self, 
        filename: str,
        mode: ModeIO = ModeIO.readwrite_binary,
        allow_deletion: bool = False,
        serializer = None,
        num_write_flush = None,
        *args, **kwargs):

        self._line_map = []
        self._io_writes = 0
        self.serializer = serializer or LazyIOPickle.serializer
        self._num_write_flush = num_write_flush or LazyIOPickle.num_write_flush
        super(LazyIOPickle, self).__init__(filename, mode, allow_deletion, *args, **kwargs)
    
    def _write(self, data, flush=False, dump_kwargs={}, *args, **kwargs):
        if self.mode.value not in ['wb', 'rb+']:
            self.setmode(mode=ModeIO.readwrite_binary)
        #self.serializer.dump(data, self._io, **dump_kwargs)
        self._io.write(self.serializer.dumps(data, **dump_kwargs))
        self._io_writes += 1
        if flush or (self._io_writes % self._num_write_flush == 0):
            self.flush()
    
    def _read(self, load_kwargs={}, *args, **kwargs):
        return self.serializer.load(self._io, **load_kwargs)
    
    def close(self):
        super(LazyIOPickle, self).close()
        self._io_writes = 0
    
    def _build_line_mapping(self):
        self._line_map.append(0)
        while self._io.readline():
            self._line_map.append(self._io.tell())
    
    def __len__(self):
        if not self._line_map:
            self._build_line_mapping()
        return len(self._line_map)

    @timed_cache(120)
    def get_num_lines(self):
        return self.__len__

    @timed_cache(120)
    def __getitem__(self, idx):
        if not self._line_map:
            self._build_line_mapping()
        self._f.seek(self._line_map[idx])
        return self.serializer.loads(self._f.readline(), ignore_errors=True, recursive=True)
    

LazyIOType = TypeVar("LazyIOType", str, LazyIOBase, LazyIOText, LazyIOJson, LazyIOJsonLines, LazyIOPickle)
LazyIOPathLike = TypeVar("LazyIOPathLike", Union[str, gfile], Union[str, os.PathLike])

class LazyIOMultiFile:    
    def __init__(
        self,
        filenames: List[LazyIOType],
        mode: ModeIO = ModeIO.read,
        ):
        self._multi_io = []
        for filename in filenames:
            if isinstance(filename, str):
                pass
        pass

# from transformers import AutoModelForSeq2SeqLM
# model_dir = 'gs://xyz/path/'
# pretrained_model = 't5-base'

# loader = LazyHFModel(AutoModelForSeq2SeqLM, pretrained_model, model_dir, model_name=None, is_base_dir=True)
# loader.load_model()
# model = loader.model


class LazyHFModel:
    def __init__(self, model_cls, base_model: str, model_dir: Optional[str] = None, model_name: Optional[str] = None, is_base_dir: Optional[bool] = True, **kwargs):
        self.model_cls = model_cls
        self.base_model = base_model
        self.base_dir = model_dir
        self.is_base_dir = is_base_dir
        if model_dir:
            if not model_name: model_name = base_model.split('/', 1)[-1].strip()
            self.model_dir = model_dir if not self.is_base_dir else File.join(model_dir, model_name)
            File.makedirs(self.model_dir)
        else:
            self.model_dir = None
        self._config = None
        self._model = None
        self._model_config_file = None
        self._model_ckpt_file = None
        self.kwargs = kwargs
        self.logger = get_logger('LazyIO', module='HFModel')
        self._testlib()

    @property
    def model_ckpt(self):
        if not self.model_dir:
            return None
        if not self._model_ckpt_file:
            for ckpt in ['model.bin', 'pytorch_model.bin']:
                model_ckpt = File.join(self.model_dir, ckpt)
                if File.exists(model_ckpt):
                    self._model_ckpt_file = model_ckpt
                    return self._model_ckpt_file
            self._model_ckpt_file = File.join(self.model_dir, 'pytorch_model.bin')
        return self._model_ckpt_file

    @property
    def model_exists(self):
        return self.model_ckpt and File.exists(self.model_ckpt)

    @property
    def model_config_file(self):
        if not self.model_dir:
            return None
        if not self._model_config_file:
            for fname in ['config.json', 'model_config.json', 'model.json']:
                config_path = File.join(self.model_dir, fname)
                if File.exists(config_path):
                    self._model_config_file = config_path
                    return self._model_config_file
        self._model_config_file = File.join(self.model_dir, 'config.json')
        return self._model_config_file


    @property
    def model(self):
        if self._model is None:
            self.load_model()
        return self._model
    
    @property
    def config(self):
        if self._config is None:
            from transformers import AutoConfig
            cfg_cls = AutoConfig.from_pretrained(self.base_model)
            if self.model_exists and self.model_config_file:
                self._config = cfg_cls.from_dict(File.jsonload(self.model_config_file))
            else:
                self._config = cfg_cls
        return self._config

    def save_model(self, model_name=None, save_config=True, state_dict=None):
        # Wrapper function around HF Save
        from transformers.modeling_utils import unwrap_model, get_parameter_dtype
        model_ckpt = File.mod_fname(self.model_ckpt, model_name, ext='bin') if model_name else self.model_ckpt
        model_to_save = unwrap_model(self._model)
        dtype = get_parameter_dtype(model_to_save)
        model_to_save.config.torch_dtype = str(dtype).split(".")[1]

        # Attach architecture to the config
        model_to_save.config.architectures = [model_to_save.__class__.__name__]
        if save_config:
            File.jsondump(model_to_save.config.to_dict(), self.model_config_file)
        
        if state_dict is None:
            state_dict = model_to_save.state_dict()

        if self._model._keys_to_ignore_on_save is not None:
            state_dict = {k: v for k, v in state_dict.items() if k not in self._model._keys_to_ignore_on_save}
        
        File.torchsave(state_dict, model_ckpt)
        self.logger.info(f'Saved Model: {model_to_save.__class__.__name__} to {model_ckpt}')


    def load_model(self, cache=True, device=None):
        device = device or self.get_device()
        if self.model_exists:
            state_dict = File.torchload(self.model_ckpt, device=device)
            self._model = self.model_cls.from_pretrained(pretrained_model_name_or_path=None, state_dict=state_dict, config=self.config, **self.kwargs)
            self.logger.info(f'Loaded Model: {self._model.__class__.__name__} from Ckpt: {self.model_ckpt}')
        else:
            self._model = self.model_cls.from_pretrained(pretrained_model_name_or_path=self.base_model, config=self.config, **self.kwargs)
            self.logger.info(f'Loaded Model: {self._model.__class__.__name__} from Pretrained: {self.base_model}')
            if cache:
                self.save_model()
    
    def load_pipeline(self, task_name='sentiment-analysis', tokenizer=None, device='auto'):
        from transformers import AutoTokenizer, pipeline
        device_id = device
        if device == 'auto':
            device_id = -1 if self.get_device().type == 'cpu' else 0
        if tokenizer:
            tokenizer = AutoTokenizer.from_pretrained(tokenizer) if isinstance(tokenizer, str) else tokenizer
        else:
            tokenizer = AutoTokenizer.from_pretrained(self.base_model)
        return pipeline(task=task_name, model=self.model, tokenizer=tokenizer, device=device_id)
        
    def _testlib(self):
        try:
            import transformers
        except ImportError:
            raise ValueError

    def reset(self):
        import torch
        torch.cuda.empty_cache()
        self._config = None
        self._model = None
        self._model_config_file = None
        self._model_ckpt_file = None

    @classmethod
    def get_device(cls):
        from fileio import torchdevice
        return torchdevice
    

'''

        # tokenizers is a mess so not supporting.
        #self._init_tokenzer = init_tokenizer
        #self._tokenizer = None
        #self._tokenizer_file = None
    @property
    def tokenizer(self):
        return self._tokenizer
@property
def tokenizer_file(self):
    if not self.model_dir:
        return None
    if not self._tokenizer_file:
        for tknzer in ['tokenizer.json', 'vocab.json']:
            tknzer_file = File.join(self.model_dir, tknzer)
            if File.exists(tknzer_file):
                self._tokenizer_file = tknzer_file
                return self._tokenizer_file
        self._tokenizer_file = File.join(self.model_dir, 'tokenizer.json')
    return self._tokenizer_file

@property
def tokenizer_exists(self):
    return self.model_ckpt and File.exists(self.tokenizer_file)
'''